/*! 
  \file editlist.cpp
  \brief Handling different edit list formats

  \author R. Enders

  $Revision: 1.26 $
  $Id: editlist.cpp,v 1.26 2006/01/06 12:17:58 reinhard Exp $
  
  Copyright (C) 2004-2006 by Reinhard Enders
  Reinhard.Enders@gmx.de

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  http://www.gnu.org/copyleft/gpl.html

  \include editlist.cpp
*/


#include "config.h"

#include "assert.h"
#include "parameter.h"
#include "editlist.h"
#include "gopdit.h"

using std::string;
using std::max;
using std::swap;

Parameter<int>  point_adj("EditPointAdj",0,
			  "adjust editpoint by so many milliseconds");
Parameter<bool> all_edit_points("DisplayCommercials",false);
Parameter<bool> import_mpeg("ImportMpeg",false,
			    "open mpegs during edit list importing");
Parameter<unsigned> suppress_clip("SuppressClip",0,
				  "suppress clips smaller than this "
				  "many seconds");

// keep editlist.cpp free of fileparam
FileDialogParameter export_file("ExportFile",
				"MTV (*.mtv *.txt);;RTV (*.rtv *.txt);;"
				"RAW (byte offsets) (*.rcl)",
				"MTV (*.mtv *.txt)",
				"Export As",
				".");

FileDialogParameter import_file("ImportFile",
				"MTV (*.mtv *.txt);;RTV (*.rtv *.txt);;"
				"RAW (byte offsets) (*.rcl)",
				"MTV (*.mtv *.txt)",
				"Import",
				".");

Parameter<int>  round_a("rounda",0,"adjust punch in editpoint");
Parameter<int>  round_d("roundd",0,"adjust punch out editpoint");
Parameter<float> framerate("framerate",0.0,
			   "framerate for converting time points "
			   "to frame numbers and vice versa");


ClipPoint ClipPointSearcher::getLastClip() const 
{ 
    return ms2offset(INT_MAX); 
}


unsigned ClipPointSearcher::offset2ms(off_t offset) const
{
    return pesmgr->getTime(offset);
}


ClipPoint ClipPointSearcher::ms2offset(unsigned ms) const
{
    // a small adjustment:
    int  adj = point_adj.getValue();
    ms = (adj > 0 || ms > unsigned(-adj)) ? ms + adj : 0;

    off_t gtop = pesmgr->getFsize();
    off_t gbot = 0;
    off_t top  = gtop;
    off_t bot  = gbot;
    off_t mid  = top / 2;
    unsigned midms = 0;

    while ( top > bot ) 
    {
	mid = ( bot + top ) / 2;
//	fprintf(stderr,"%llu %llu %llu ",bot,mid,top);

	mid = pesmgr->getNextGop(mid);
//	fprintf(stderr,"%llu\n",mid);

// might not work with small intervals
	
	if ( (mid >= top && top == gtop) || mid == off_t(-1) ) 
	{
	    return ClipPoint(gtop,midms);
	}

	midms = pesmgr->getTime(mid);

	if ( ms == midms ) 
	{
	    return ClipPoint(mid,midms);
	}
	else if ( midms > ms ) 
	{
	    if ( mid >= top ) 
	    {
		top = (bot + top) / 2;
	    }
	    else 
	    {
		top = mid;
	    }
	}
	else if ( midms < ms ) 
	{
	    if ( bot == mid ) 
	    {
		break;
	    }
	    bot = mid;
	}
    }
    return ClipPoint(mid,midms);
}



AbstractEditListMgr::AbstractEditListMgr(FILE* pfp, Gopdit& gd, GopditUi& pui)
    : ClipPointSearcher(gd.getPesMgr())
    , fp(pfp)
    , gopdit(gd)
    , ui(pui)
{ 
}



/*! 
  EVENTDUMP format:
  The first character is the entry type:
  F - file name of the source mpg (from an RTV set, .mpg/.evt/.ndx)
  A - add point time (punch in)
  D - delete point time (punch out)
  E - end editing

  Time is in MMM:SS.ms  (Minutes, Seconds, milliseconds)

Also supported by RTVEDIT:
  T - target for new file set (optional)
 */
class RtvEditListMgr : public AbstractEditListMgr
{
public:
    RtvEditListMgr(FILE* pfp, Gopdit& gopdit, GopditUi& ui, bool pframes)
	: AbstractEditListMgr(pfp, gopdit, ui)
	, frames(pframes)
    {
	frame_rate = (framerate.getValue() > 0.0)
	    ? framerate.getValue()
	    : (( pesmgr != 0 ) ? pesmgr->getFrameRate() : 30);
    }
    virtual Mpeg2ClipList importlist();
    virtual void exportlist(const Mpeg2ClipList& clip_list);    
protected:
    void    trace(char s[], ClipPoint p)
    {
	char buf[256];
	sprintf(buf,"%s frame_rate: %.2f ms: %d frames: %d offset: %llu",
		s, frame_rate, p.ms, ms2frames(p.ms), p.off);
	ui.report(string(buf));
    }

    int     ms2frames(int ms) const { 
	return int((ms / 1000.0) * frame_rate +  0.5); }
    int     frames2ms(int fs) const { 
	return int((fs / frame_rate) * 1000 + 0.5); }
    // if true frames otherwise millisecons
    ClipPoint adjustpoint(ClipPoint point, unsigned ms, int adjust);
    ClipPoint nextGopPoint(ClipPoint point);
    ClipPoint prevGopPoint(ClipPoint point);

    bool    frames;
    float   frame_rate;
};


ClipPoint RtvEditListMgr::nextGopPoint(ClipPoint point)
{
    off_t noff = pesmgr->getNextGop(point.off + 1);
    return ClipPoint(noff, pesmgr->getTime(noff));
}


ClipPoint RtvEditListMgr::prevGopPoint(ClipPoint point)
{
    off_t noff = pesmgr->getPrevGop(point.off);
    return ClipPoint(noff, pesmgr->getTime(noff));
}


ClipPoint RtvEditListMgr::adjustpoint(ClipPoint point, unsigned ms, int adjust)
{
    ClipPoint npoint;
    // look for closest point
    if ( adjust == 0 ) 
    {
	if ( ms < point.ms ) 
	{
//	    fprintf(stderr,"prevGopPoint\n");
	    npoint = prevGopPoint(point);
	}
	else if ( ms > point.ms ) 
	{
//	    fprintf(stderr,"nextGopPoint\n");
	    npoint = nextGopPoint(point);
	}
	else 
	{
	    return point;
	}
//	trace("altern pnt",npoint);
	return ( abs(point.ms - ms) <= abs(npoint.ms - ms) ) 
	    ? point : npoint;
    }
    // look for the closest point with fuzzy (adjust - 1)
    else if ( adjust > 0 ) 
    {
	npoint = prevGopPoint(point);
	if ( npoint.ms >= ms - frames2ms(adjust-1)  ) 
	{
	    return npoint;
	}
	else if ( point.ms >= ms - frames2ms(adjust-1) ) 
	{
	    return point;
	}
	return nextGopPoint(point);
    }
    else // adjust < 0
    {
	npoint = nextGopPoint(point);
	if ( npoint.ms <= ms + frames2ms(-adjust-1)  ) 
	{
	    return npoint;
	}
	else if ( point.ms <= ms + frames2ms(-adjust-1) ) 
	{
	    return point;
	}
	return prevGopPoint(point);
    }
    return point;
}


Mpeg2ClipList RtvEditListMgr::importlist()
{
    Mpeg2ClipList clip_list;
    
    Mpeg2Clip   tclip;
    char c;
    tclip.pesmgr = pesmgr;
    bool open_clip = false;

    while ( (c = fgetc(fp)) != EOF ) 
    {
	int ms;
	unsigned sec;
	unsigned min;
	unsigned framenr;
	char filename[256];

	switch ( c ) 
	{
	case 'E':
	    break;
	case 'A':
	    open_clip = true;
	    if ( pesmgr == 0 ) 
	    {
		ui.error("No Movie Loaded");
		return clip_list;
	    }
	    if ( frames ) 
	    {
		fscanf(fp,"%d", &framenr);
		ms = frames2ms(framenr);
	    }
	    else 
	    {
		fscanf(fp,"%d:%d.%d",&min,&sec,&ms);
		ms += ((min * 60) + sec) * 1000;
	    }
	    ms = max(0,ms);
	    fprintf(stderr, "\n");
	    trace("search   A", ClipPoint(0,ms));
	    if ( all_edit_points.getValue() ) 
	    {
		tclip.end = ms2offset(ms);
		tclip.mode = Mpeg2Clip::Commercial;
		if ( tclip.end.ms - tclip.begin.ms > suppress_clip.getValue() ) 
		{
		    clip_list.push_back(tclip);
		    tclip.begin = tclip.end;
		    tclip.mode = Mpeg2Clip::Show;
		}
	    }
	    else 
	    {
		tclip.begin  = ms2offset(ms);
		trace("found    A", tclip.begin);
		tclip.begin = adjustpoint(tclip.begin, ms, round_a.getValue());
		trace("adjusted A", tclip.begin);
	    }
	    break;
	case 'D':
	    open_clip = false;
	    if ( pesmgr == 0 ) 
	    {
		ui.error("No Movie Loaded");
		return clip_list;
	    }
	    if ( frames ) 
	    {
		fscanf(fp,"%d", &framenr);
		ms = frames2ms(framenr);
	    }
	    else 
	    {
		fscanf(fp,"%d:%d.%d",&min,&sec,&ms);
		ms += ((min * 60) + sec) * 1000;
	    }
	    ms = max(0,ms);
	    fprintf(stderr, "\n");
	    trace("search   D", ClipPoint(0,ms));
	    tclip.end = ms2offset(ms);
	    trace("found    D", tclip.end);
	    tclip.end = adjustpoint(tclip.end, ms, round_d.getValue());
	    trace("adjusted D", tclip.end);

	    clip_list.push_back(tclip);
	    if ( all_edit_points.getValue() ) 
	    {
		tclip.begin = tclip.end;
	    }
	    break;
	case 'F':
	    if ( !import_mpeg.getValue() ) 
	    {
		break;
	    }
	    if ( all_edit_points.getValue() || open_clip ) 
	    {
		tclip.end = ms2offset(INT_MAX);
		clip_list.push_back(tclip);
	    }
	    open_clip = false;
	    fscanf(fp,"%[^\n]",filename);
	    if ( !gopdit.openFile(string(filename)) ) 
	    {
		return clip_list;
	    }
	    pesmgr = gopdit.getPesMgr();
	    // initialze a bit, in order to get framerate etc.
	    pesmgr->play(1,true);
	    frame_rate = (framerate.getValue() > 0.0)
		? framerate.getValue()
		: pesmgr->getFrameRate();
	    tclip.pesmgr = pesmgr;
	    tclip.begin = ms2offset(0);

	    break;
	case 'S':
	    fscanf(fp,"%[^\n]",filename);
	    tclip.filename = filename;
	    break;
	default:
	    break;
	}
	while ( (c = fgetc(fp)) != EOF && c != '\n' ) 
	{}
    }
    if ( all_edit_points.getValue() || open_clip ) 
    {
	tclip.end = ms2offset(INT_MAX);
	clip_list.push_back(tclip);
    }
    return clip_list;
}

void RtvEditListMgr::exportlist(const Mpeg2ClipList& clip_list)
{
    pesmgr = 0;

    // print clips
    for ( unsigned i = 0; i < clip_list.size(); ++i ) 
    {
	if ( clip_list[i].pesmgr != pesmgr ) 
	{
	    // print file name
	    pesmgr = clip_list[i].pesmgr;
	    fprintf(fp,"F%s\n",pesmgr->getFileName().c_str());
	}

	if ( clip_list[i].filename != string() ) 
	{
	    fprintf(fp,"S%s\n",clip_list[i].filename.c_str());
	}

	unsigned ms;
	unsigned min;
	unsigned sec;

	ms = offset2ms(clip_list[i].begin.off);
	if ( frames ) 
	{
	    fprintf(fp,"A%08d\n", ms2frames(ms));
	}
	else
	{
	    min = ms / 60000;
	    ms -= min * 60000;
	    sec = ms / 1000;
	    ms -= sec * 1000;
	    fprintf(fp,"A%03d:%02d.%03d\n",min,sec,ms);
	}

	ms = offset2ms(clip_list[i].end.off);
	if ( frames ) 
	{
	    fprintf(fp,"D%08d\n", ms2frames(ms));
	}
	else
	{
	    min = ms / 60000;
	    ms -= min * 60000;
	    sec = ms / 1000;
	    ms -= sec * 1000;
	    fprintf(fp,"D%03d:%02d.%03d\n",min,sec,ms);	    
	}
    }
    // print end marker
    fprintf(fp,"E\n");
}



//! exporting / importing byte pure byte offsets (mainly for testing)
class RawEditListMgr : public AbstractEditListMgr
{
public:
    RawEditListMgr(FILE* pfp, Gopdit& gopdit, GopditUi& ui)
	: AbstractEditListMgr(pfp, gopdit, ui)
    {}
    virtual Mpeg2ClipList importlist();
    virtual void exportlist(const Mpeg2ClipList& clip_list);    
};

Mpeg2ClipList RawEditListMgr::importlist()
{
    Mpeg2ClipList clip_list;
    
    Mpeg2Clip   tclip;

    if ( pesmgr == 0 ) 
    {
	ui.error("No Movie Loaded");
	return clip_list;
    }


    tclip.pesmgr = pesmgr;
    
    while ( true ) 
    {
	if ( 2 > fscanf(fp,"%llu %llu",&tclip.begin.off,&tclip.end.off) ) 
	{
	    break;
	}
	tclip.begin.ms = offset2ms(tclip.begin.off);
	tclip.end.ms   = offset2ms(tclip.end.off);
	clip_list.push_back(tclip);
    }
    return clip_list;
}

void RawEditListMgr::exportlist(const Mpeg2ClipList& clip_list)
{
    for ( unsigned i = 0; i < clip_list.size(); ++i ) 
    {
	fprintf(fp,"%llu %llu\n",clip_list[i].begin.off, clip_list[i].end.off);
    }
}

void GopditEditListMgr::exportList(string filename, string filter, 
				   const Mpeg2ClipList& clip_list)
{
    FILE* out;
    out = fopen(filename.c_str(),"wb");
    
    if ( out == 0 ) 
    {
	ui.error("Cannot open file " + filename);
	return;
    }
    
    AbstractEditListMgr* elm = genEditListMgr(filter, out);
    if ( elm != 0 ) 
    {
	elm->exportlist(clip_list);
	delete elm;
    }
    fclose(out);
}

Mpeg2ClipList GopditEditListMgr::importList(string filename, string filter)
{
    Mpeg2ClipList clip_list;

    FILE* in;

    if ( filename != "-" ) 
    {
	in = fopen(filename.c_str(),"r");
	
	if ( in == 0 )
	{
	    ui.error("Cannot open file " + filename);
	    return clip_list;
	}
    }
    else 
    {
	in = stdin;
    }
    AbstractEditListMgr* elm = genEditListMgr(filter, in);
    if ( elm != 0 ) 
    {
	clip_list = elm->importlist();
	delete elm;
    }
    if ( in != stdin ) 
    {
	fclose(in);
    }

    return clip_list;
}

// should ultimately be a map,
AbstractEditListMgr* GopditEditListMgr::genEditListMgr(string filter, FILE* fp)
{
    string filetype = string(filter,0,filter.find(" "));

    if ( filetype == "MTV" ) 
    {
	return new RtvEditListMgr(fp, gopdit, ui, true);
    }
    if ( filetype == "RTV" ) 
    {
	return new RtvEditListMgr(fp, gopdit, ui, false);
    }
    if ( filetype == "RAW" ) 
    {
	return new RawEditListMgr(fp, gopdit, ui);
    }
    else 
    {
	ui.error("Unknown Editlist Type");
	return 0;
    }
}
