#ifndef PESMGR_H
#define PESMGR_H

/*! 
  \file pesmgr.h

  \brief Manager class that cares about PES streams

  \author R. Enders

  $Revision: 1.20 $
  $Id: pesmgr.h,v 1.20 2006/01/06 12:18:00 reinhard Exp $
  
  Copyright (C) 2004-2006 by Reinhard Enders
  Reinhard.Enders@gmx.de

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  http://www.gnu.org/copyleft/gpl.html

  \include pesmgr.h
*/


// A wrapper class for pesmgr
// initialization and deinitialization

#include <inttypes.h>

#include <string>
#include <vector>

#include "gopditui.h"
#include "pesfilter.h"

extern "C" {
#include <mpeg2dec/mpeg2.h>
}

class PesMgr;

struct ClipPoint
{
    ClipPoint()
	: off(0)
	, ms(0)
    {}
    ClipPoint(off_t o, unsigned m)
	: off(o)
	, ms(m)
    {}
    
    //! byte offset
    off_t    off;
    //! milliseconds
    unsigned ms;
};

struct Mpeg2Clip
{
    Mpeg2Clip()
	: pesmgr(0)
	, mode(Show)
    {}
    enum Mode { Show, Commercial };
    PesMgr*         pesmgr;
    ClipPoint       begin;
    ClipPoint       end;
    Mode            mode;
    std::string      filename;
};

typedef std::vector<Mpeg2Clip> Mpeg2ClipList;

#include <map>

#define BUFFER_SIZE 4096
#define BUFFER_PREFIX 8

class PesAbstractDecoder;
class GopditOutput;

class PesMgr
{
public:
    PesMgr(GopditUi& pui, GopditOutput& poutput,
	   const char* pfilename, const char* moviename);

    ~PesMgr();

    bool         concat(off_t& sb, FILE* out, off_t from, off_t to,
			unsigned& frame_no, float fps)
    {
	return psmgr->concat(sb, out,from,to,frame_no,fps);
    }
    void         endOfProgram(FILE* out)
    {
	psmgr->endOfProgram(out);
    }

    // true, if painting was successful
    bool         draw();

    void         resize(int width, int height);

    void         findNextGop();

    void         findPrevGop(unsigned n);

    off_t        getCurGop() const;

    off_t        getValidCurGop() const;

    unsigned     getCurTime() const { return cur_time; }

    //! get the next gop (starting from offset)
    off_t        getNextGop(off_t offset);

    off_t        getPrevGop(off_t cur_gop);

    //! get the time of gop (in milliseconds)
    unsigned     getTime(off_t gop_offset);

    std::string  getMovieName() const { return moviename; }

    //! true, if current gop is valid
    bool         isValidCurGop() const; 

    bool         isValidOffset(off_t offset) const 
    { return offset != off_t(-1); }

    // if safe is true play anyway and always display frames
    void         play(long gops, bool safe = false);

    void         playOneSlice(bool safe = false);

    void         setCurTime(unsigned t) { cur_time = t; }

    std::string  getFileName() const { return psmgr->getFileName(); }

    off_t        getFsize() const { return psmgr->getFsize(); }

    ClipPoint    getLastClipPoint();

    void         dumpInfo();

    time_t       getMtime() const { return psmgr->getMtime(); }

    bool         setOffset(off_t off);

    bool         isRunning() const { return running; }
    void         setRunning(bool r) { running = r; }
        
    mpeg2dec_t&  getParser() const { return *parser; }
    void         reinitParser();

    GopditUi&    getUi() const { return ui; }

    bool         getSequenceInfoValid() const { return sequence_info_valid; }
    void         setSequenceInfo(const mpeg2_sequence_s& si)
    { 
	sequence_info_valid = true;
	sequence_info = si; 
    }
    float        getFrameRate() const;
    int          getBitRate() const;
    int          getPWidth() const;
    int          getPHeight() const;
private:
    // some more methods

    inline size_t incBuffer();

    void         abstractDecode(PesAbstractDecoder& m2d);

    // some more control state and variables
    off_t        offset;
    bool         running;
    // current gop offset
    off_t        gopoff;
    // the maximum number of bytes we had to back for finding the previous gop
    off_t        backoff;

    // some more variables
    PesFilter*   psmgr;

    std::string  moviename;

    // current time in milliseconds
    unsigned     cur_time;

    bool         sequence_info_valid;
    mpeg2_sequence_s sequence_info;
    mpeg2dec_t*  parser;
    uint8_t      pbuffer[BUFFER_SIZE+BUFFER_PREFIX];
    uint8_t*     buffer;
    size_t       size;

    // caching gop predecessors
    std::map<off_t,off_t> prev_cache;

    GopditOutput& output;

    // well - may be should be api
    GopditUi&    ui;
};


#endif /* PESMGR_H */
